# flake8: noqa

from langchain.prompts import PromptTemplate
from langchain.retrievers.multi_query import LineListOutputParser

from .verbatims import VerbatimsFromDoc, verbatims_parser, Verbatims

_map_verbatim_parser = LineListOutputParser()

_question_prompt_template = """
Use the following portion of a long document to see if any of the text is relevant to answer the question.
---
{context}
---

Question: {question}

Extract all verbatims from texts relevant to answering the question in separate strings else output an empty array.
{format_instructions}

"""

QUESTION_PROMPT = PromptTemplate(
    template=_question_prompt_template,
    input_variables=["context", "question"],
    partial_variables={
        "format_instructions": _map_verbatim_parser.get_format_instructions()
    },
    output_parser=_map_verbatim_parser,
)

# _map_verbatim_parser: BaseOutputParser = PydanticOutputParser(
#     pydantic_object=VerbatimsFromDoc
# )

# Use some object. It's easier to update the verbatims schema.
_response_example_1 = Verbatims(
    response="This Agreement is governed by English law.",
    documents=[
        VerbatimsFromDoc(
            # ids=["_idx_0"],
            ids=[""],
            verbatims=[
                "This Agreement is governed by English law",
                # "The english law is applicable for this agreement."
            ],
        ),
        VerbatimsFromDoc(
            # ids=["_idx_4"],
            ids=[""],
            verbatims=["The english law is applicable for this agreement."],
        ),
    ],
)

_response_example_2 = Verbatims(response="", documents=[])

_sample_1 = VerbatimsFromDoc(
    # ids=["_idx_0", "_idx_1"],
    ids=[],
    verbatims=["This Agreement is governed by English law"],
)
_sample_2 = VerbatimsFromDoc(
    # ids=["_idx_2"],
    ids=[],
    verbatims=[],
)
_sample_3 = VerbatimsFromDoc(
    # ids=["_idx_3"],
    ids=[],
    verbatims=[],
)
_sample_4 = VerbatimsFromDoc(
    # ids=["_idx_4"],
    ids=[],
    verbatims=["The english law is applicable for this agreement."],
)

_sample_5 = VerbatimsFromDoc(
    # ids=["_idx_0"],
    ids=[],
    verbatims=[
        "Madam Speaker, Madam Vice President, our First Lady and Second "
        "Gentleman. Members of Congress and the Cabinet."
    ],
)
_sample_6 = VerbatimsFromDoc(
    # ids=["_idx_1", "_idx_2"],
    ids=[],
    verbatims=[],
)
_sample_7 = VerbatimsFromDoc(
    # ids=["_idx_3"],
    ids=[],
    verbatims=[],
)
_sample_8 = VerbatimsFromDoc(
    # ids=["_idx_4"],
    ids=[],
    verbatims=[],
)

# _question_prompt_template = """
# Use the following portion of a long document to see if any of the text is relevant to answer the question.
# ---
# {context}
# ---
#
# Question: {question}
#
# Extract all verbatims from texts relevant to answering the question in separate strings else output an empty array.
# {format_instructions}
#
# """
#
# QUESTION_PROMPT = PromptTemplate(
#     template=_question_prompt_template,
#     input_variables=["context", "question"],
#     partial_variables={
#         "format_instructions": _map_verbatim_parser.get_format_instructions()
#     },
#     output_parser=_map_verbatim_parser,
# )

_combine_prompt_template = """Given the following extracts from several documents, 
a question and not prior knowledge. 

Process step by step:
- extract all verbatims
- extract all associated ids
- create a final response with these verbatims
- produces the json result

QUESTION: Which state/country's law governs the interpretation of the contract?
=========
Content: {_sample_1}
Content: {_sample_2}
Content: {_sample_3}
Content: {_sample_4}
=========
FINAL ANSWER:
{response_example_1}

QUESTION: What did the president say about Michael Jackson?
=========
Content: {_sample_5}
Content: {_sample_6}
Content: {_sample_7}
Content: {_sample_8}
=========
FINAL ANSWER: 
{response_example_2}

QUESTION: {question}
=========
{summaries}
=========
If you are not confident with your answer, say 'I don't know'. 
{format_instructions}
FINAL ANSWER:"""
COMBINE_PROMPT = PromptTemplate(
    template=_combine_prompt_template,
    input_variables=["summaries", "question"],
    partial_variables={
        "format_instructions": verbatims_parser.get_format_instructions(),
        "response_example_1": _response_example_1.json(),
        "response_example_2": _response_example_2.json(),
        "_sample_1": _sample_1.json(),
        "_sample_2": _sample_2.json(),
        "_sample_3": _sample_3.json(),
        "_sample_4": _sample_4.json(),
        "_sample_5": _sample_5.json(),
        "_sample_6": _sample_6.json(),
        "_sample_7": _sample_7.json(),
        "_sample_8": _sample_8.json(),
    },
    output_parser=verbatims_parser,
)

EXAMPLE_PROMPT = PromptTemplate(
    template="Ids: {_idx}\n" "Content: {page_content}\n",
    input_variables=["page_content", "_idx"],
)



# flake8: noqa
from langchain.prompts import PromptTemplate
from .verbatims import verbatims_parser

_initial_qa_prompt_template = """
Context information is below. 
---------------------
{context_str}
---------------------
Process step by step:
- ignore prior knowledge
- extract references ("IDX")
- extract all the exact verbatims from the texts only if they are relevant to answering the question, in a list of strings 
- create a final answer
- produce the json result

Given the context information the question: {question}

If you don't know the answer, just say that you don't know, without references, and does not extract any verbatim. 
Don't try to make up an answer.

The ids must be only in the form '_idx_<number>'.
{format_instructions}
"""

INITIAL_QA_PROMPT = PromptTemplate(
    input_variables=["context_str", "question"],
    template=_initial_qa_prompt_template,
    partial_variables={
        "format_instructions": verbatims_parser.get_format_instructions()
    },
    output_parser=verbatims_parser,
)

_refine_prompt_template = """
Given the context information and not prior knowledge answer, the question: {question}

We have provided an existing JSON answer with the list of documents with associated ids and verbatims of texts: 
{existing_answer}

We have the opportunity to refine the existing answer (only if needed) with some more context below.
------------
{context_str}
------------
Given the new context, refine the original answer to better answer the question. 
If you don't know how to refine the original answer, does not modify the answer.

Process step by step:
- ignore prior knowledge
- with the new context 
    - extract references of the new context ("IDX")
    - add all the exact verbatims from the texts of the new context, only if they are relevant to answering the question, in a list of strings 
- refine the original answer to better answer the question. ONLY if you do update it
- append the new IDS and add verbatims of texts from the existing answser IDS as well. 
- produce the result

ALWAYS return a "IDS" part in your answer. 
If the context isn't useful, return the original answer.

If you don't know the answer, just say that you don't know. Don't try to make up an answer.

{format_instructions}
"""

REFINE_PROMPT = PromptTemplate(
    input_variables=["question", "existing_answer", "context_str"],
    template=_refine_prompt_template,
    partial_variables={
        "format_instructions": verbatims_parser.get_format_instructions()
    },
    output_parser=verbatims_parser,
)


EXAMPLE_PROMPT = PromptTemplate(
    template="ids: {_idx}\n" "Content: {page_content}\n",
    input_variables=["page_content", "_idx"],
)



"""Question answering with references over documents."""

import inspect
import logging
import re
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Set, Tuple, cast

from langchain.base_language import BaseLanguageModel
from langchain.callbacks.manager import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
)
from langchain.chains.base import Chain
from langchain.chains.combine_documents.base import BaseCombineDocumentsChain
from langchain.chains.combine_documents.map_reduce import MapReduceDocumentsChain
from langchain.chains.combine_documents.reduce import ReduceDocumentsChain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.chains.llm import LLMChain
from langchain.docstore.document import Document
from langchain.pydantic_v1 import Extra
from langchain.schema import BaseOutputParser, BasePromptTemplate

from .loading import (
    load_qa_with_references_chain,
)
from .map_reduce_prompts import (
    COMBINE_PROMPT,
    EXAMPLE_PROMPT,
    QUESTION_PROMPT,
)

logger = logging.getLogger(__name__)


# TOTRY: Add in VectorStoreIndexWrapper.query_with_references()
# TOTRY: Add in VectorStoreIndexWrapper.query_with_references_and_verbatims()
class BaseQAWithReferencesChain(Chain, ABC):
    combine_documents_chain: BaseCombineDocumentsChain

    """Chain to use to combine documents."""
    question_key: str = "question"  #: :meta private:
    input_docs_key: str = "docs"  #: :meta private:
    answer_key: str = "answer"  #: :meta private:
    source_documents_key: str = "source_documents"  #: :meta private:
    chain_type: str  #: :meta private:
    output_parser: Optional[BaseOutputParser] = None

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        document_prompt: BasePromptTemplate = EXAMPLE_PROMPT,
        question_prompt: BasePromptTemplate = QUESTION_PROMPT,
        combine_prompt: BasePromptTemplate = COMBINE_PROMPT,
        **kwargs: Any,
    ) -> "BaseQAWithReferencesChain":
        """Construct the chain from an LLM."""
        llm_question_chain = LLMChain(llm=llm, prompt=question_prompt)
        llm_combine_chain = LLMChain(llm=llm, prompt=combine_prompt)
        combine_results_chain = StuffDocumentsChain(
            llm_chain=llm_combine_chain,
            document_prompt=document_prompt,
            document_variable_name="summaries",
        )
        reduce_documents_chain = ReduceDocumentsChain(
            combine_documents_chain=combine_results_chain
        )
        combine_document_chain = MapReduceDocumentsChain(
            llm_chain=llm_question_chain,
            reduce_documents_chain=reduce_documents_chain,
            document_variable_name="context",
            return_intermediate_steps=True,
        )
        return cls(
            combine_documents_chain=combine_document_chain,
            **kwargs,
        )

    @classmethod
    def from_chain_type(
        cls,
        llm: BaseLanguageModel,
        chain_type: str = "stuff",
        chain_type_kwargs: Optional[dict] = None,
        **kwargs: Any,
    ) -> "BaseQAWithReferencesChain":
        """Load chain from chain type."""
        _chain_kwargs = chain_type_kwargs or {}
        combine_document_chain = load_qa_with_references_chain(
            llm,
            chain_type=chain_type,
            **_chain_kwargs,
        )
        if chain_type == "map_rerank" and "output_parser" not in kwargs:
            from .map_rerank_prompts import rerank_reference_parser

            kwargs["output_parser"] = rerank_reference_parser

        return cls(
            combine_documents_chain=combine_document_chain,
            chain_type=chain_type,
            **kwargs,
        )

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    @property
    def input_keys(self) -> List[str]:
        """Expect input key.

        :meta private:
        """
        return [self.question_key]

    @property
    def output_keys(self) -> List[str]:
        """Return output key.

        :meta private:
        """
        _output_keys = [self.answer_key, self.source_documents_key]
        return _output_keys

    def _split_sources(self, answer: str) -> Tuple[str, str]:
        """Split sources from answer."""
        if re.search(r"SOURCES?:", answer, re.IGNORECASE):
            answer, sources = re.split(
                r"SOURCES?:|QUESTION:\s", answer, flags=re.IGNORECASE
            )[:2]
            sources = re.split(r"\n", sources)[0].strip()
        else:
            sources = ""
        return answer, sources

    @abstractmethod
    def _get_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: CallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""

    def _process_reference(
        self, answers: Dict[str, Any], docs: List[Document], references: Any
    ) -> Set[int]:
        # With the map_rerank mode, use extra parameter for map_rerank to identify
        # the corresponding document.
        if "_idx" in answers:
            references.documents_ids.add(answers["_idx"])

        documents_idx = set()
        for str_doc_id in references.documents_ids:
            m = re.match(r"_idx_(\d+)", str_doc_id.strip())
            if m:
                documents_idx.add(int(m[1]))
            else:
                pass
        return documents_idx

    def _process_results(
        self,
        answers: Dict[str, Any],
        docs: List[Document],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Tuple[str, Set[int]]:
        idx = set()
        answer = answers[self.combine_documents_chain.output_key]
        # At this time (version 0.0.273), the output_parser is not
        # automatically called.
        # We must extract this parser to analyse the response.
        parser: Optional[BaseOutputParser] = self.output_parser
        llm_chain: Optional[LLMChain] = None
        if not parser:
            if self.chain_type == "map_rerank":
                assert self.output_parser
            elif self.chain_type == "map_reduce":
                llm_chain = cast(
                    Any, self.combine_documents_chain
                ).collapse_document_chain.llm_chain
            elif self.chain_type == "refine":
                llm_chain = cast(Any, self.combine_documents_chain).refine_llm_chain
            else:
                # Simple chain
                llm_chain = cast(Any, self.combine_documents_chain).llm_chain
        if llm_chain:
            parser = llm_chain.prompt.output_parser
        assert parser

        references = parser.parse(answer)
        answer = references.response

        idx = self._process_reference(answers, docs, references)

        for doc in docs:
            del doc.metadata["_idx"]
        return answer, idx
        # except OutputParserException as e:
        #     # Probably that the answer has been cut off.
        #     raise e
        # except Exception as e:
        #     if run_manager:
        #         run_manager.on_chain_error(e)
        #     raise e

    @abstractmethod
    async def _aget_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: AsyncCallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        accepts_run_manager = (
            "run_manager" in inspect.signature(self._get_docs).parameters
        )
        if accepts_run_manager:
            docs = self._get_docs(inputs, run_manager=_run_manager)
        else:
            docs = self._get_docs(inputs)  # type: ignore[call-arg]

        # Inject position in the list
        # To avoid confusion with other id, add a prefix
        for idx, doc in enumerate(docs):
            doc.metadata["_idx"] = f"_idx_{idx}"

        answers = self.combine_documents_chain(
            {
                self.combine_documents_chain.input_key: docs,
                self.question_key: inputs[self.question_key],
                **inputs,
            },
            callbacks=_run_manager.get_child(),
        )
        answer, all_idx = self._process_results(answers, docs)
        selected_docs = [docs[idx] for idx in all_idx if idx < len(docs)]

        return {
            self.answer_key: answer,
            self.source_documents_key: selected_docs,
        }

    # TODO: aligner le code
    async def _acall(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        _run_manager = run_manager or AsyncCallbackManagerForChainRun.get_noop_manager()
        accepts_run_manager = (
            "run_manager" in inspect.signature(self._aget_docs).parameters
        )
        if accepts_run_manager:
            docs = await self._aget_docs(inputs, run_manager=_run_manager)
        else:
            docs = await self._aget_docs(inputs)  # type: ignore[call-arg]

        # Inject position in the list
        for idx, doc in enumerate(docs):
            doc.metadata["_idx"] = idx

        await self.combine_documents_chain.arun(
            input_documents=docs, callbacks=_run_manager.get_child(), **inputs
        )
        answers = await self.combine_documents_chain.acall(
            {
                self.combine_documents_chain.input_key: docs,
                self.question_key: inputs[self.question_key],
                **inputs,
            },
            callbacks=_run_manager.get_child(),
        )
        answer, all_idx = self._process_results(answers, docs)
        selected_docs = [docs[idx] for idx in all_idx if idx < len(docs)]

        return {
            self.answer_key: answer,
            self.source_documents_key: selected_docs,
        }


class QAWithReferencesChain(BaseQAWithReferencesChain):
    """
    This chain extracts the information from the documents that was used to answer the
    question. The output `source_documents` contains only the documents that were used,
    and for each one.

    The result["source_documents"] returns only the list of documents used.
    Then it is possible to find the page of a PDF, or the chapter of a markdown.
    """

    input_docs_key: str = "docs"  #: :meta private:

    @property
    def input_keys(self) -> List[str]:
        """Expect input key.

        :meta private:
        """
        return [self.input_docs_key, self.question_key]

    def _get_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: CallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""
        return inputs.pop(self.input_docs_key)

    async def _aget_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: AsyncCallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""
        return inputs.pop(self.input_docs_key)

    @property
    def _chain_type(self) -> str:
        return "qa_with_references_chain"



# flake8: noqa


from langchain.prompts import PromptTemplate
from .verbatims import verbatims_parser, Verbatims, VerbatimsFromDoc

_response_example_1 = Verbatims(
    response="This Agreement is governed by English law.",
    documents=[
        # Sample with multiple verbatims
        VerbatimsFromDoc(
            ids=["_idx_0"],
            verbatims=[
                "This Agreement is governed by English law",
                "The english law is applicable for this agreement.",
            ],
        ),
        # Sample with one verbatim
        VerbatimsFromDoc(
            ids=["_idx_3"],
            verbatims=["The english law is applicable for this agreement."],
        ),
    ],
)

# Sample with no result
_response_example_2 = Verbatims(response="I don't know.", documents=[])

_template = """Given the following extracts from several documents, a question and not prior knowledge. 
Process step by step:
- for each documents:
    - extract the references ("IDS")
    - extract all the verbatims from the texts only if they are relevant to answering the question, in a list of strings  
- creates a final answer
- produces the json result

If you don't know the answer, just say that you don't know. Don't try to make up an answer.
ALWAYS return a "IDS" part in your answer in another line.

QUESTION: Which state/country's law governs the interpretation of the contract?
=========
Content: This Agreement is governed by English law and the parties submit to the exclusive jurisdiction of the English courts in relation to any dispute (contractual or non-contractual) concerning this Agreement save that either party may apply to any court for an injunction or other relief to protect its Intellectual Property Rights.
Ids: _idx_0

Content: No Waiver. Failure or delay in exercising any right or remedy under this Agreement shall not constitute a waiver of such (or any other) right or remedy.
Ids: _idx_1

Content: (b) if Google believes, in good faith, that the Distributor has violated or caused Google to violate any Anti-Bribery Laws
Ids: _idx_2

Content: The english law is applicable for this agreement.
Ids: _idx_3
=========
FINAL ANSWER: 
```
{_response_example_1}
```

QUESTION: What did the president say about Michael Jackson?
=========
Content: Madam Speaker, Madam Vice President, our First Lady and Second Gentleman. Members of Congress and the Cabinet.
Ids: _idx_0

Content: And we won’t stop. \n\nWe have lost so much to COVID-19. Time with one another. And worst of all, so much loss of life.
Ids: _idx_1

Content: And a proud Ukrainian people, who have known 30 years of independence, have repeatedly shown that they will not tolerate anyone who tries to take their country backwards.
Ids: _idx_2

Content: More support for patients and families. \n\nTo get there, I call on Congress to fund ARPA-H, the Advanced Research Projects Agency for Health.
ids: _idx_3
=========
FINAL ANSWER: 
```
{_response_example_2}
```

QUESTION: {question}
=========
{summaries}
=========
The ids must be only in the form '_idx_<number>'.
{format_instructions}
FINAL ANSWER: 
"""

PROMPT = PromptTemplate(
    template=_template,
    input_variables=["summaries", "question"],
    partial_variables={
        "format_instructions": verbatims_parser.get_format_instructions(),
        "_response_example_1": _response_example_1.json(),
        "_response_example_2": _response_example_2.json(),
    },
    output_parser=verbatims_parser,
)
EXAMPLE_PROMPT = PromptTemplate(
    template="Content: {page_content}\n" "idx: {_idx}",
    input_variables=["page_content", "_idx"],
)



import logging
import re
from typing import Set

from langchain.output_parsers import PydanticOutputParser
from langchain.pydantic_v1 import BaseModel, Extra
from langchain.schema import BaseOutputParser

logger = logging.getLogger(__name__)

# To optimize the consumption of tokens, it's better to use only 'text', without json.
# Else the schema consume ~300 tokens and the response 20 tokens by step
_OPTIMIZE = True  # Experimental


class References(BaseModel):
    """
    Response and referenced documents.
    """

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    response: str
    """ The response """
    documents_ids: Set[str] = set()
    """ The list of documents used to response """

    def __str__(self) -> str:
        if _OPTIMIZE:
            return f'{self.response}\nIDX:{",".join(map(str, self.documents_ids))}'
        else:
            return self.json()


references_parser: BaseOutputParser
if _OPTIMIZE:

    class _ReferencesParser(BaseOutputParser):
        """An optimised parser for Reference.
        It's more effective than the pydantic approach
        """

        @property
        def lc_serializable(self) -> bool:
            return True

        @property
        def _type(self) -> str:
            """Return the type key."""
            return "reference_parser"

        def get_format_instructions(self) -> str:
            return (
                "Your response should be in the form:\n"
                "Answer:the response\n"
                "IDX: a comma-separated list of document identifiers used "
                "in the response"
            )

        def parse(self, text: str) -> References:
            regex = r"(?:Answer:)?(.*)\nIDX:(.*)"
            match = re.search(regex, text)
            if match:
                ids: Set[int] = set()
                for str_doc_id in match[2].split(","):
                    m = re.match(r"(?:_idx_)?(\d+)", str_doc_id.strip())
                    if m:
                        ids.add(int(m[1]))

                return References(response=match[1].strip(), documents_ids=ids)
            else:
                raise ValueError(f"Could not parse output: {text}")

    references_parser = _ReferencesParser()
else:
    references_parser = PydanticOutputParser(pydantic_object=References)
